@@ -17,9 +17,11 @@
 
 from .kml import KML, Document, Folder, Placemark
 from .kml import TimeSpan, TimeStamp
+from .kml import ExtendedData, Data
+from .kml import Schema, SchemaData
 
 from .styles import StyleUrl, Style, StyleMap
 from .styles import IconStyle, LineStyle, PolyStyle
-from .styles import LabelStyle #, BalloonStyle
+from .styles import LabelStyle, BalloonStyle
 
 from .atom import Link, Author, Contributor
