@@ -29,6 +29,7 @@ try:
     import urlparse
 except ImportError:
     import urllib.parse as urlparse
+import warnings
 
 from .geometry import Point, LineString, Polygon
 from .geometry import MultiPoint, MultiLineString, MultiPolygon
@@ -48,7 +49,7 @@ logger = logging.getLogger('fastkml.kml')
 
 from .config import etree
 
-from .base import _BaseObject
+from .base import _BaseObject, _XMLObject
 
 from .styles import StyleUrl, Style, StyleMap, _StyleSelector
 
@@ -110,14 +111,17 @@ class KML(object):
             raise TypeError
 
     def etree_element(self):
-        root = etree.Element('%skml' % self.ns)
-
         # self.ns may be empty, which leads to unprefixed kml elements.
         # However, in this case the xlmns should still be mentioned on the kml
         # element, just without prefix.
         if not self.ns:
+            root = etree.Element('%skml' % self.ns)
             root.set('xmlns', config.NS[1:-1])
-
+        else:
+            if config.LXML:
+                root = etree.Element('%skml' % self.ns, nsmap={None:self.ns[1:-1]})
+            else:
+                root = etree.Element('%skml' % self.ns)
         for feature in self.features():
             root.append(feature.etree_element())
         return root
@@ -386,6 +390,38 @@ class _Feature(_BaseObject):
                 yield style
             else:
                 raise TypeError
+    @property
+    def snippet(self):
+        if self._snippet:
+            if isinstance(self._snippet, dict):
+                text = self._snippet.get('text')
+                if text:
+                    assert (isinstance(text, basestring))
+                    max_lines = self._snippet.get('maxLines', None)
+                    if max_lines is None:
+                        return {'text': text}
+                    elif int(max_lines) > 0:
+                        # if maxLines <=0 ignore it
+                        return {'text': text, 'maxLines': max_lines}
+            elif isinstance(self._snippet, basestring):
+                return self._snippet
+            else:
+                raise ValueError("Snippet must be dict of {'text':t, 'maxLines':i} or string")
+
+    @snippet.setter
+    def snippet(self, snip=None):
+        self._snippet = {}
+        if isinstance(snip, dict):
+            self._snippet['text'] = snip.get('text')
+            max_lines = snip.get('maxLines')
+            if max_lines is not None:
+                self._snippet['maxLines'] = int(snip['maxLines'])
+        elif isinstance(snip, basestring):
+            self._snippet['text'] = snip
+        elif snip is None:
+            self._snippet = None
+        else:
+            raise ValueError("Snippet must be dict of {'text':t, 'maxLines':i} or string")
 
     def etree_element(self):
         element = super(_Feature, self).etree_element()
@@ -404,15 +440,15 @@ class _Feature(_BaseObject):
             element.append(self._styleUrl.etree_element())
         for style in self.styles():
             element.append(style.etree_element())
-        if self._snippet:
+        if self.snippet:
             snippet = etree.SubElement(element, "%sSnippet" %self.ns)
-            if isinstance(self._snippet, basestring):
-                snippet.text = self._snippet
+            if isinstance(self.snippet, basestring):
+                snippet.text = self.snippet
             else:
-                assert (isinstance(self._snippet['text'], basestring))
-                snippet.text = self._snippet['text']
-                if self._snippet.get('maxLines'):
-                    snippet.set('maxLines', str(self._snippet['maxLines']))
+                assert (isinstance(self.snippet['text'], basestring))
+                snippet.text = self.snippet['text']
+                if self.snippet.get('maxLines'):
+                    snippet.set('maxLines', str(self.snippet['maxLines']))
         if (self._time_span is not None) and (self._time_stamp is not None):
             raise ValueError('Either Timestamp or Timespan can be defined, not both')
         elif self._time_span is not None:
@@ -459,9 +495,10 @@ class _Feature(_BaseObject):
             self._styleUrl = s
         snippet = element.find('%sSnippet' % self.ns)
         if snippet is not None:
-            self._snippet = {'text': snippet.text}
+            _snippet = {'text': snippet.text}
             if snippet.get('maxLines'):
-                self._snippet['maxLines'] = int(snippet.get('maxLines'))
+                _snippet['maxLines'] = int(snippet.get('maxLines'))
+            self.snippet = _snippet
         timespan = element.find('%sTimeSpan' % self.ns)
         if timespan is not None:
             s = TimeSpan(self.ns)
@@ -484,14 +521,13 @@ class _Feature(_BaseObject):
             self._atom_author = s
         extended_data = element.find('%sExtendedData' % self.ns)
         if extended_data is not None:
-            if extended_data.find('%sData' % self.ns):
-                x = UntypedExtendedData(self.ns)
-                x.from_element(extended_data)
-                self.extended_data = x
-            else:
-                logger.warn(
-                    'arbitrary or typed extended data is not yet supported'
-                )
+            x = ExtendedData(self.ns)
+            x.from_element(extended_data)
+            self.extended_data = x
+            #else:
+            #    logger.warn(
+            #        'arbitrary or typed extended data is not yet supported'
+            #    )
 
 
 
@@ -543,9 +579,25 @@ class _Container(_Feature):
 class Document(_Container):
     """
     A Document is a container for features and styles. This element is
-    required if your KML file uses shared styles
+    required if your KML file uses shared styles or schemata for typed
+    extended data
     """
     __name__ = "Document"
+    _schemata = None
+
+    def schemata(self):
+        if self._schemata:
+            for schema in self._schemata:
+                yield schema
+
+    def append_schema(self, schema):
+        if self._schemata is None:
+            self._schemata = []
+        if isinstance(schema, Schema):
+            self._schemata.append(schema)
+        else:
+            s = Schema(schema)
+            self._schemata.append(s)
 
     def from_element(self, element):
         super(Document, self).from_element(element)
@@ -559,6 +611,18 @@ class Document(_Container):
             feature = Placemark(self.ns)
             feature.from_element(placemark)
             self.append(feature)
+        schemata = element.findall('%sSchema' % self.ns)
+        for schema in schemata:
+            s = Schema(self.ns, id = 'default')
+            s.from_element(schema)
+            self.append_schema(s)
+
+    def etree_element(self):
+        element = super(Document, self).etree_element()
+        if self._schemata is not None:
+            for schema in self._schemata:
+                element.append(schema.etree_element())
+        return element
 
     def get_style_by_url(self, styleUrl):
         id = urlparse.urlparse(styleUrl).fragment
@@ -799,7 +863,117 @@ class TimeSpan(_TimePrimitive):
         return element
 
 
-class UntypedExtendedData(_BaseObject):
+class Schema(_BaseObject):
+    """
+    Specifies a custom KML schema that is used to add custom data to
+    KML Features.
+    The "id" attribute is required and must be unique within the KML file.
+    <Schema> is always a child of <Document>.
+    """
+    __name__ = "Schema"
+
+    _simple_fields = None
+    # The declaration of the custom fields, each of which must specify both the
+    # type and the name of this field. If either the type or the name is
+    # omitted, the field is ignored.
+    name = None
+
+    def __init__(self, ns=None, id=None, name=None, fields=None):
+        if id is None:
+            raise ValueError('Id is required for schema')
+        super(Schema, self).__init__(ns, id)
+        self.simple_fields = fields
+        self.name = name
+
+    @property
+    def simple_fields(self):
+        sfs = []
+        for simple_field in self._simple_fields:
+            if simple_field.get('type') and simple_field.get('name'):
+                sfs.append( {'type': simple_field['type'],
+                    'name': simple_field['name'],
+                    'displayName': simple_field.get('displayName')})
+        return tuple(sfs)
+
+    @simple_fields.setter
+    def simple_fields(self, fields):
+        self._simple_fields = []
+        if isinstance(fields, dict):
+            self.append(**fields)
+        elif isinstance(fields, (list, tuple)):
+            for field in fields:
+                if isinstance(field, (list, tuple)):
+                    self.append(*field)
+                elif isinstance(field, dict):
+                    self.append(**field)
+        elif fields is None:
+            self._simple_fields = []
+        else:
+            raise ValueError('Fields must be of type list, tuple or dict')
+
+    def append(self, type, name, displayName=None):
+        """
+        append a field.
+        The declaration of the custom field, must specify both the type
+        and the name of this field.
+        If either the type or the name is omitted, the field is ignored.
+
+        The type can be one of the following:
+            string
+            int
+            uint
+            short
+            ushort
+            float
+            double
+            bool
+
+        <displayName>
+        The name, if any, to be used when the field name is displayed to
+        the Google Earth user. Use the [CDATA] element to escape standard
+        HTML markup.
+        """
+        allowed_types= ['string', 'int', 'uint', 'short', 'ushort',
+                        'float', 'double', 'bool']
+        if type not in allowed_types:
+            raise TypeError("type must be one of 'string', 'int', 'uint', 'short', 'ushort', 'float', 'double', 'bool'")
+        else:
+            #TODO explicit type conversion to check for the right type
+            pass
+        self._simple_fields.append({'type': type, 'name': name,
+                                    'displayName': displayName})
+
+    def from_element(self, element):
+        super(Schema, self).from_element(element)
+        self.name = element.get('name')
+        simple_fields = element.findall('%sSimpleField' % self.ns)
+        self.simple_fields = None
+        for simple_field in simple_fields:
+            sfname = simple_field.get('name')
+            sftype = simple_field.get('type')
+            display_name = simple_field.find('%sdisplayName' % self.ns)
+            if display_name is not None:
+                sfdisplay_name = display_name.text
+            else:
+                sfdisplay_name = None
+            self.append(sftype, sfname, sfdisplay_name)
+
+    def etree_element(self):
+        element = super(Schema, self).etree_element()
+        if self.name:
+            element.set('name', self.name)
+        for simple_field in self.simple_fields:
+            sf = etree.SubElement(element, "%sSimpleField" %self.ns)
+            sf.set('type', simple_field['type'])
+            sf.set('name', simple_field['name'])
+            if simple_field.get('displayName'):
+                dn = etree.SubElement(sf, "%sdisplayName" %self.ns)
+                dn.text = simple_field['displayName']
+
+        return element
+
+
+class ExtendedData(_XMLObject):
     """ Represents a list of untyped name/value pairs. See docs:
 
     -> 'Adding Untyped Name/Value Pairs'
@@ -808,61 +982,136 @@ class UntypedExtendedData(_BaseObject):
     """
     __name__ = 'ExtendedData'
 
-    def __init__(self, ns=None, id=None, elements=None):
-        super(UntypedExtendedData, self).__init__(ns, id)
+    def __init__(self, ns=None, elements=None):
+        super(ExtendedData, self).__init__(ns)
         self.elements = elements or []
 
     def etree_element(self):
-        element = super(UntypedExtendedData, self).etree_element()
-
+        element = super(ExtendedData, self).etree_element()
         for subelement in self.elements:
             element.append(subelement.etree_element())
-
         return element
 
     def from_element(self, element):
-        super(UntypedExtendedData, self).from_element(element)
+        super(ExtendedData, self).from_element(element)
         self.elements = []
-
-        for subelement in element:
-            el = UntypedExtendedDataElement(self.ns)
-            el.from_element(subelement)
-
+        untyped_data = element.findall('%sData' % self.ns)
+        for ud in untyped_data:
+            el = Data(self.ns)
+            el.from_element(ud)
+            self.elements.append(el)
+        typed_data = element.findall('%sSchemaData' % self.ns)
+        for sd in typed_data:
+            el = SchemaData(self.ns, 'dummy')
+            el.from_element(sd)
             self.elements.append(el)
 
 
-class UntypedExtendedDataElement(_BaseObject):
+class UntypedExtendedData(ExtendedData):
+
+    def __init__(self, ns=None, elements=None):
+        super(UntypedExtendedData, self).__init__(ns, elements)
+        warnings.warn("UntypedExtendedData is deprecated use ExtendedData instead", DeprecationWarning)
+
+
+class Data(_XMLObject):
     """ Represents an untyped name/value pair with optional display name. """
 
     __name__ = 'Data'
 
-    def __init__(self, ns=None, id=None, name=None, value=None, display_name=None):
-        super(UntypedExtendedDataElement, self).__init__(ns, id)
+    def __init__(self, ns=None, name=None, value=None, display_name=None):
+        super(Data, self).__init__(ns)
 
         self.name = name
         self.value = value
         self.display_name = display_name
 
     def etree_element(self):
-        element = super(UntypedExtendedDataElement, self).etree_element()
-
+        element = super(Data, self).etree_element()
         element.set('name', self.name)
-
         value = etree.SubElement(element, "%svalue" % self.ns)
         value.text = self.value
-
         if self.display_name:
             display_name = etree.SubElement(element, "%sdisplayName" % self.ns)
             display_name.text = self.display_name
-
         return element
 
     def from_element(self, element):
-        super(UntypedExtendedDataElement, self).from_element(element)
-
+        super(Data, self).from_element(element)
         self.name = element.get('name')
         self.value = element.find('%svalue' % self.ns).text
-
         display_name = element.find('%sdisplayName' % self.ns)
         if display_name is not None:
             self.display_name = display_name.text
+
+class UntypedExtendedDataElement(Data):
+    def __init__(self, ns=None, name=None, value=None, display_name=None):
+        super(UntypedExtendedDataElement, self).__init__(ns, name, value, display_name)
+        warnings.warn("UntypedExtendedDataElement is deprecated use Data instead", DeprecationWarning)
+
+class SchemaData(_XMLObject):
+    """
+    <SchemaData schemaUrl="anyURI">
+    This element is used in conjunction with <Schema> to add typed
+    custom data to a KML Feature. The Schema element (identified by the
+    schemaUrl attribute) declares the custom data type. The actual data
+    objects ("instances" of the custom data) are defined using the
+    SchemaData element.
+    The <schemaURL> can be a full URL, a reference to a Schema ID defined
+    in an external KML file, or a reference to a Schema ID defined
+    in the same KML file.
+    """
+    __name__ = 'SchemaData'
+    schema_url = None
+    _data = None
+
+    def __init__(self, ns=None, schema_url=None, data=None):
+        super(SchemaData, self).__init__(ns)
+        if (not isinstance(schema_url, basestring)) or (not schema_url):
+            raise ValueError('required parameter schema_url missing')
+        self.schema_url = schema_url
+        self._data = []
+        self.data = data
+
+    @property
+    def data(self):
+        return tuple(self._data)
+
+    @data.setter
+    def data(self, data):
+        if isinstance(data, (tuple, list)):
+            self._data = []
+            for d in data:
+                if isinstance(d, (tuple, list)):
+                    self.append_data(*d)
+                elif isinstance(d, dict):
+                    self.append_data(**d)
+        elif data is None:
+            self._data = []
+        else:
+            raise TypeError('data must be of type tuple or list')
+
+    def append_data(self, name, value):
+        if isinstance(name, basestring) and name:
+            self._data.append({'name':name, 'value': value})
+        else:
+            raise TypeError('name must be a nonempty string')
+
+    def etree_element(self):
+        element = super(SchemaData, self).etree_element()
+        element.set('schemaUrl', self.schema_url)
+        for data in self.data:
+            sd = etree.SubElement(element, "%sSimpleData" %self.ns)
+            sd.set('name', data['name'])
+            sd.text = data['value']
+        return element
+
+    def from_element(self, element):
+        super(SchemaData, self).from_element(element)
+        self.data = []
+        self.schema_url = element.get('schemaUrl')
+        simple_data = element.findall('%sSimpleData' % self.ns)
+        for sd in simple_data:
+            self.append_data(sd.get('name'), sd.text)
+
+
